require(igraph)

####################################
### Custom functions
####################################

## Generate a total graph of the landscape
total.graph<-function(n.property.sides)
{
  n.parcels<-(n.property.sides*2)^2
  vertex.attrs <- list(name=1:n.parcels)#list(name = unique(c(df$src, df$dst)))
  edges <- cbind(rep(1:n.parcels, times=2), rep(NA, n.parcels*2))
  edges[1:n.parcels,2]<-edges[1:n.parcels,1]+1 # down
  edges[(n.parcels+1):(2*n.parcels-1),2]<-edges[(n.parcels+1):(2*n.parcels-1),1]+n.property.sides*2 # to the left
  edges<-edges[-(1:(n.property.sides*2))*n.property.sides*2,] #remove down
  edges<-edges[-which(edges[,2]>n.parcels),]; edges<-edges[-which(is.na(edges[,2])),]
  
  G <- graph.empty(n = 0, directed = FALSE)
  G <- add.vertices(G, length(vertex.attrs$name), attr = vertex.attrs)
  G[from=edges[,1], to=edges[,2]]<-TRUE
  #G <- add.edges(G, edges[1:2,])
  return(G)
}

## Determine if two parcels are neighbours
is.neighbour<-function(x1,y1,x2,y2)
{
  return(abs(x1-x2)<=1)&&(abs(y1-y2)<=1)
}

## Find all neighbours of a given parcel
find.neighbours<-function(x,y,n.sides)
{
  neighbours.x<-c(x-1,x,x+1)
  neighbours.y<-c(y-1,y,y+1)
  x.neighbours<-expand.grid(x, neighbours.y)  # if want to include diagonals, neighbours<-expand.grid(neightbours.x,neighbours.y)
  y.neighbours<-expand.grid(neighbours.x, y)
  neighbours<-rbind(x.neighbours, y.neighbours)
  neighbours<-unique(neighbours)
  neighbours<-neighbours[which((neighbours[,1]%in%1:n.sides)&(neighbours[,2]%in%1:n.sides)),]
  neighbours<-neighbours[-which((neighbours[,1]==x)*(neighbours[,2]==y)==1),]
  return(neighbours)
}

## Convert between index and xy coordinates of parcel
coords.to.index<-function(Row,Col,n.sides)
{
  index<-Row+(Col-1)*n.sides
  return(index)
}
index.to.coords<-function(index, n.sides)
{
  x<-index%%n.sides; x[which(x==0)]<-n.sides
  y<-ceiling(index/n.sides)
  
  return(matrix(c(x,y),ncol=2))
}

## create adjacency matrix for input into igraph -- only native habitat are nodes, and links are between spatial neighbours
create.adj<-function(M, n.sides)
{
  adj<-matrix(0, n.sides^2, n.sides^2)
  adj[row(adj)==col(adj)][which(M==1)]<-1 #set self-adjacency
  for (pctr in which(M==1)) #cycle through only those parcels with conserved habitat
  {
    # ALL neighbours of that parcal
    neighbours<-find.neighbours(index.to.coords(pctr, n.sides)[1,1],index.to.coords(pctr, n.sides)[1,2], n.sides)
    for (nctr in 1:nrow(neighbours))
    {
      n<-coords.to.index(neighbours[nctr,1], neighbours[nctr,2], n.sides)
      if (n>=1 && n<=n.parcels) #don't go out of bounds
      {
        adj[pctr,n]<-adj[n,n] # only adjacent if the second parcel is also conserved habitat
      }
    }
  }
  return(adj)
}

## Convert landscape into graph
make.graph<-function(M)
{
  n.sides<-nrow(M)
  adj<-matrix(NA,n.sides, n.sides)
  adj<-create.adj(M, n.sides);g<-graph.adjacency(adj)
  return(g)
}

find.edge<-function(M)
{
  n.sides<-nrow(M)
  nearest.edge<-nearest.edge<-array(NA,dim=c(n.sides,n.sides))
  for (pctr in which(M==1)) # cycle through only those parcels with conserved habitat
  {
    # find coordinates
    point1<-index.to.coords(pctr, n.sides); x1<-point1[1,1]; y1<-point1[1,2]
    # find coordinates of all 0s
    all.zeros<-index.to.coords(which(M==0),n.sides)
    # min(abs(x1-x0)) = shortest x distance
    x.diff<-min(abs(all.zeros[,1]-x1))
    # min(abs(y1-y0)) = shortest y distance
    y.diff<-min(abs(all.zeros[,2]-y1))
    # use shortest of those two
    nearest.edge[x1,y1]<-min(x.diff,y.diff)
  }
  return(nearest.edge)
}

########################
## Remove parcels that can't/won't be changed from the potential action set
## This might be because 
##    - property is already exactly compliant (therefore won't change)
##    - Parcels are native habitat on property that is undercompliant (future clearing is banned)
##    - Parcels are agriculture on overcompliant property (don't need to change)
########################

lock.in<-function(conserved, ag.prod, target, locked.in)
{
  n.sides<-dim(conserved)[1]
  n.property.sides<-n.sides/2
  n.properties<-n.property.sides^2
  n.parcels<-n.sides^2
  
  ####################################
  ### Lock untradeable parcels in
  ####################################
  
  #  locked.in<-array(0,dim=c(n.sides,n.sides))
  
  prop.locked<-prop.comp<-prop.cons<-rep(NA,n.properties)
  under.profit<-over.profit<-matrix(NA,n.properties,2)
  n.allowable.trades<-rep(0, n.properties)
  # determining property compliance (aim 50%)
  for (prop.ctr in 1:n.properties)
  {
    # describe property by top left-hand corner
    property.coords<-index.to.coords(prop.ctr,n.property.sides)
    prop.x<-property.coords[1,1]*2-1
    prop.y<-property.coords[1,2]*2-1
    property<-expand.grid((0:1)+prop.x, (0:1)+prop.y)
    # determine property compliance
    prop.cons[prop.ctr]<-mean(conserved[0:1+prop.x,0:1+prop.y])
    if (any(locked.in[0:1+prop.x,0:1+prop.y])!=1)
    {    
      if (prop.cons[prop.ctr]==target) 
      {
        # lock in properties that are exactly compliant
        locked.in[0:1+prop.x,0:1+prop.y]<-1
        prop.comp[prop.ctr]<-prop.locked[prop.ctr]<-1
      }else if (prop.cons[prop.ctr]<target){ 
        # lock in the conserved land on undercompliant properties
        locked.in[0:1+prop.x,0:1+prop.y]<-conserved[0:1+prop.x,0:1+prop.y] # binary matrix -- 0 entries are available for optimization
        prop.profit<-ag.prod[0:1+prop.x,0:1+prop.y]*(1-conserved[0:1+prop.x,0:1+prop.y]);prop.profit[prop.profit==0]<-NA # profit in property
        n.allowable.trades[prop.ctr]<-2-4*prop.cons[prop.ctr]
        for (i in 1:min(length(prop.profit),n.allowable.trades[prop.ctr]))
        {
          under.profit[prop.ctr,i]<-sort(prop.profit, decreasing=FALSE)[i]
        }
      }else if ((prop.cons[prop.ctr]>target)&(prop.cons[prop.ctr]<1)){ 
        # lock in the converted land on overcompliant properties
        locked.in[0:1+prop.x,0:1+prop.y]<-1-conserved[0:1+prop.x,0:1+prop.y] # binary matrix -- 0 entries are available for optimization
        prop.comp[prop.ctr]<-1
        prop.profit<-ag.prod[0:1+prop.x,0:1+prop.y]*(conserved[0:1+prop.x,0:1+prop.y]);prop.profit[prop.profit==0]<-NA #profit in property
        n.allowable.trades[prop.ctr]<-1
        for (i in 1:min(length(prop.profit),n.allowable.trades[prop.ctr]))
        {
          over.profit[prop.ctr,i]<-sort(prop.profit, decreasing=TRUE)[i]
        }
      }else if (prop.cons[prop.ctr]==1){
        prop.profit<-ag.prod[0:1+prop.x,0:1+prop.y]*(conserved[0:1+prop.x,0:1+prop.y]);prop.profit[prop.profit==0]<-NA #profit in property
        n.allowable.trades[prop.ctr]<-2
        for (i in 1:min(length(prop.profit),n.allowable.trades[prop.ctr]))
        {
          over.profit[prop.ctr,i]<-sort(prop.profit, decreasing=TRUE)[i]
        }
      }
    } 
  }
  to.allocate<-1-locked.in; to.allocate[to.allocate==0]<-NA
  allocated<-locked.in;allocated[allocated==0]<-NA
  ill.all<-allocated;ill.all[ill.all==1]<-'X'
  return(list(to.allocate=to.allocate, allocated=allocated, ill.all=ill.all, under.profit=under.profit, 
              over.profit=over.profit, n.allowable.trades=n.allowable.trades, locked.in=locked.in))
}

##############################
## Trade model
###############################

run.market<-function(conserved.new, under.profit, over.profit, ag.prod,ill.all,clearag,reforest, n.allowable.trades)
{
  n.sides<-dim(conserved.new)[1]
  n.parcels<-n.sides^2
  trade<-array(NA,dim=c(n.sides, n.sides, n.habs))
  i<-1
  must.restore<-numeric()
  wctr<-0
  while (sum(!is.na(under.profit))>0&&(wctr<5000)) # there are still potential buyers
  {
    wctr<-wctr+1
    if (max(under.profit,na.rm=T)>=min(over.profit,na.rm=T))#min(c(under.profit[,1],over.profit[,1]), na.rm=T)%in%over.profit[,1])
    {
      # can afford trade  
      
      # initial prices
      prices<-over.profit[which(!is.na(over.profit))]
      offers<-under.profit[which(!is.na(under.profit))]
      
      prices<-c(sort(prices[which(prices<=max(offers, na.rm=T))]),NA)
      offers<-c(sort(offers[which(offers>=min(prices,na.rm=T))]),NA)
      
      price<-min(prices, na.rm=T)
      next.price<-max(min(prices[-which.min(prices)]), min(offers, na.rm=T), na.rm=T) # workaround that sets the "next price" as the smallest offer if there is only one price
      
      buyers<-offers[which(offers>price & offers<=next.price)]
      buyer<-max(buyers)
      
      trade[which(ag.prod==price)]<- -i # record sale
      ill.all[which(ag.prod==price)]<-paste("T ", -i, round(price,2))
      ill.all[which(ag.prod==buyer)]<-paste("T ", i, round(buyer,2))
      trade[which(ag.prod==buyer)]<-i # record purchase
      
      prop<-index.to.coords(which(over.profit==price),length(n.allowable.trades))[1,1]
      n.allowable.trades[prop]<-n.allowable.trades[prop]-1
      over.profit<-over.profit*(n.allowable.trades>0);over.profit[over.profit==0]<-NA
      prop<-index.to.coords(which(under.profit==buyer),length(n.allowable.trades))[1,1]
      n.allowable.trades[prop]<-n.allowable.trades[prop]-1
      under.profit<-under.profit*(n.allowable.trades>0);under.profit[under.profit==0]<-NA 
      over.profit[which(over.profit==price)]<-under.profit[which(under.profit==buyer)]<-NA
      if (sum(!is.na(under.profit))+(sum(!is.na(over.profit)))>0)
      {
        if (sum(!is.na(under.profit))*(sum(!is.na(over.profit)))==0) #one has run out
        {
          clearag[which(ag.prod==over.profit[which(!is.na(over.profit))])]<-1
          conserved.new[which(ag.prod==over.profit[which(!is.na(over.profit))])]<-0
          reforest[which(ag.prod==under.profit[which(!is.na(under.profit))])]<-1
          conserved.new[which(ag.prod==under.profit[which(!is.na(under.profit))])]<-1
          over.profit[which(!is.na(over.profit))]<-NA
          under.profit[which(!is.na(under.profit))]<-NA
        }
        # force remaining LUCs for those that can't afford
        temp<-forced.LUC(conserved.new, ag.prod, under.profit, over.profit, ill.all, clearag, reforest, n.allowable.trades)
        over.profit<-temp$over.profit; under.profit<-temp$under.profit; ill.all<-temp$ill.all;
        conserved.new<-temp$conserved.new; clearag<-temp$clearag; reforest<-temp$reforest
      }
      
      i<-i+1
    }else{
      # force remaining LUCs for those that can't afford
      temp<-forced.LUC(conserved.new, ag.prod, under.profit, max.prop.pot,profit, ill.all, clearag, reforest, n.allowable.trades)
      over.profit<-temp$over.profit; under.profit<-temp$under.profit; ill.all<-temp$ill.all;
      conserved.new<-temp$conserved.new; clearag<-temp$clearag; reforest<-temp$reforest
      n.allowable.trades<-temp$n.allowable.trades
    }
  }
  return(list(conserved=conserved, conserved.new=conserved.new, ill.all=ill.all, reforest=reforest, clearag=clearag, trade=trade))
}

find.changes<-function(habs.cons1, conserved1, habs.cons.new, conserved.new, ag.prod, n.habs, n.property.sides)
{
  n.sides<-dim(habs.cons1)[1]
  n.parcels<-n.sides^2
  
  # monetary
  tot.prod.initial<-sum((1-conserved1)*ag.prod)   
  tot.prod.final<-sum((1-conserved.new)*ag.prod)
  prod.change<-tot.prod.final-tot.prod.initial
  prod.prop.change<-prod.change/tot.prod.initial
  
  ####################################
  ### Convert landscape to graph
  ####################################

  # graph theory investigation into landscape
  
  PC.change<-mean.csize.change<-median.csize.change<-min.csize.change<-max.csize.change<-rep(NA,h)
  cnum.change<-ea.change<-hab.rep.change<-mean.dist.to.edge.change<-rep(NA,h)
  G<-total.graph(n.property.sides) # this is a custom function -- takes n.property sides
  for (h in 1:n.habs)
  {
#    print(h)
#    print(habs.cons1)
#    plot(G)
    
#    summary(G)
    # convert landscapes to graphs
    init.g<-delete.vertices(G, which(habs.cons1[,,h]==0))
    final.g<-delete.vertices(G, which(habs.cons.new[,,h]==0))
    
    ####################################
    ### BENEFIT: Calculating "landscape benefit" as probability of connectivity
    ###################################
    
    # Calculating the Probability of connectivity
    # for simple grid, easy to calc: PC = number of parcels with connections/total number of parcels in landscape squred
    init.PC<- sum(clusters(init.g)$csize^2)/(n.parcels^2) # dividing it is just to make it a probability
    final.PC <- sum(clusters(final.g)$csize^2)/(n.parcels^2) # dividing it is just to make it a probability
    
    ## Other metrics
    cs<-clusters(init.g)
    init.g.csize<-cs$csize #sizes of subgraphs
    init.g.cnum<-cs$no #number of subgraphs
    init.ea.ratio<-(vcount(init.g)*4-2*ecount(init.g))/vcount(init.g) # "edges" of forest = nodes*4-edges (graph edges are where there are not forest "edges". not confusing terminology at all.)
    
    cs<-clusters(final.g)
    final.g.csize<-cs$csize #sizes of subgraphs
    final.g.cnum<-cs$no #number of subgraphs
    final.ea.ratio<-(vcount(final.g)*4-2*ecount(final.g))/vcount(final.g) # "edges" of forest = nodes*4-edges (graph edges are where there are not forest "edges". not confusing terminology at all.)
    
    init.edge<-mean(find.edge(habs.cons1[,,h]), na.rm=TRUE)
    final.edge<-mean(find.edge(habs.cons.new[,,h]), na.rm=TRUE)
    mean.dist.to.edge.change[h]<-(final.edge-init.edge)/max(init.edge,1)
    
    # these are all proportions
    PC.change[h]<-(final.PC-init.PC)/init.PC
    mean.csize.change[h]<-(mean(final.g.csize)-mean(init.g.csize))/mean(init.g.csize) #mean fragment size
    median.csize.change[h]<-(median(final.g.csize)-median(init.g.csize))/median(init.g.csize)
    min.csize.change[h]<-(min(final.g.csize)-min(init.g.csize))/min(init.g.csize) #min fragment size
    max.csize.change[h]<-(max(final.g.csize)-max(init.g.csize))/max(init.g.csize) #max fragment size
    cnum.change[h]<-(final.g.cnum-init.g.cnum)/init.g.cnum # number of fragments
    ea.change[h]<-(final.ea.ratio-init.ea.ratio)/init.ea.ratio # edge to area ratio (higher is better)
    hab.rep.change[h]<-(sum(habs.cons.new[,,h])-sum(habs.cons1[,,h]))/sum(habs.cons1[,,h])
  }
  PC.change<-mean(PC.change)
  mean.csize.change<-mean(mean.csize.change)
  median.csize.change<-mean(median.csize.change)
  min.csize.change<-mean(min.csize.change)
  max.csize.change<-mean(max.csize.change)
  cnum.change<-mean(cnum.change)
  ea.change<-mean(ea.change)
  hab.rep.change<-min(hab.rep.change)
  mean.dist.to.edge.change<-mean(mean.dist.to.edge.change)
  
  return(list(prod.change=prod.change, prod.prop.change=prod.prop.change, PC.change=PC.change, median.csize.change=median.csize.change,
              mean.csize.change=mean.csize.change, min.csize.change=min.csize.change, max.csize.change=max.csize.change,
              cnum.change=cnum.change,ea.change=ea.change, mean.dist.to.edge.change=mean.dist.to.edge.change, min.hab.rep.change=hab.rep.change))
}


run.market.new<-function(conserved.new, under.profit, over.profit, ag.prod,ill.all,clearag,reforest, n.allowable.trades, p.deforest)
{
  n.sides<-dim(conserved.new)[1]
  n.parcels<-n.sides^2
  trade<-array(NA,dim=c(n.sides, n.sides, n.habs))
  i<-1
  must.restore<-numeric()
  
  demand<-sort(under.profit, decreasing=T)
  supply<-sort(over.profit)
  length(demand) <- length(supply) <- max(c(length(demand), length(supply)))
  price<-supply[max(which(demand>supply))]
  
  # undercompliant who can't afford price are reforested
  conserved.new[which(ag.prod%in%demand[which(demand<price)])]<-1
  reforest[which(ag.prod%in%demand[which(demand<price)])]<-1
  # can afford -- just record trade
  trade[which(ag.prod%in%demand[which(demand>=price)])]<-1
  
  # overcompliant who are too expensive for price are deforested
  temp.deforest<-rbinom(length(which(supply>price)),1,(1-p.deforest))
  conserved.new[which(ag.prod%in%supply[which(supply>price)])]<-temp.deforest#0
  clearag[which(ag.prod%in%supply[which(supply>price)])]<-1
  # can afford -- just record trade
  trade[which(ag.prod%in%supply[which(supply<=price)])]<-1
  
  return(list(conserved=conserved, conserved.new=conserved.new, ill.all=ill.all, reforest=reforest, clearag=clearag, trade=trade))
}



##########################
## Visually inspect landscape
##########################

create.Graph<-function(conserved.mat)
{
  n.sides<-dim(conserved.mat)[1]
  
  ####################################
  ### Convert landscape to graph
  ####################################
  require(igraph)
  # graph theory investigation into landscape
  
  G<-total.graph(n.sides/2) # this is a custom function -- takes n.property sides
  init.g<-delete.vertices(G, which(conserved.mat==0))
  return(init.g)
}

plot.landscape<-function(conserved.mat)
{
  gs<-create.Graph(conserved.mat)
  trade.g<-gs#$trade.g
  #      plot(trade.g, vertex.label=NA, vertex.size=1)
  #largest clique:
  #max.frag<-index.to.coords(V(trade.g)[which(cs$membership==which.max(cs$csize))]$name, n.sides)
  cs<-clusters(gs)
  conserved.mat[V(trade.g)[which(cs$membership==which.max(cs$csize))]$name]<-3
  cluster.ea.ratio<-rep(NA,clusters(trade.g)$no)
  for (i in 1:clusters(trade.g)$no)
  {
    #V(trade.g)[which(clusters(trade.g)$membership == i)]
    # delete vertices not in cluster
    cut.g<-delete.vertices(trade.g,V(trade.g)[which((V(trade.g)$name%in%V(trade.g)[which(clusters(trade.g)$membership == i)]$name)==0)])
    cluster.ea.ratio[i]<-(vcount(cut.g)*4-2*ecount(cut.g))/vcount(cut.g)
  }
  conserved.mat[V(trade.g)[which(cs$membership==which.min(cluster.ea.ratio))]$name]<-2
  heatmap.2(conserved.mat, dendrogram='none', Rowv=FALSE, Colv=FALSE, 
            labRow=' ', labCol=' ',
            #cellnote=round(ag.prod,1), notecol="black", 
            trace='none', key=FALSE,lwid = c(0.03,0.2),lhei = c(.03,.2), col = c("white", "lightgrey", "darkgrey", "black"))
}













